# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import copy
from collections import OrderedDict, defaultdict
from functools import reduce
from typing import Dict, Union

import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from ..compose import ModelStructure
from ..compose._parse import parse_encoder_definition
from ..encode import CategoricalEncoder
from ..utils import as_list, check_X, check_y
from .linear import LinearPredictor


class GroupedPredictor(RegressorMixin, BaseEstimator):
    """Construct one predictor per data group.

    The predictor splits data by the different values of a single column and fits one
    estimator per group. Since each of the models in the ensemble predicts on a different
    subset of the input data (an observation cannot belong to more than one clusters),
    the final prediction is generated by vertically concatenating all the individual
    models’ predictions.

    Args:
        group_feature (str): The name of the column of the input dataframe to use as
            the grouping set.
        model_conf (Dict[str, Dict]): A dictionary that includes information about the
            base model's structure.
        feature_conf (Dict[str, Dict], optional): A dictionary that maps feature
            generator names to the classes for the generators' validation and
            creation. Defaults to None.
        estimator_params (dict or tuple of tuples, optional): The parameters to use when
            instantiating a new base estimator. If none are given, default parameters are
            used. Defaults to tuple().
        fallback (bool, optional): Whether or not to fall back to a global model in case a
            group parameter is not found during `.predict()`. Otherwise, an exception will
            be raised. Defaults to False.
    """

    def __init__(
        self,
        *,
        group_feature: str,
        model_conf: Dict[str, Dict],
        feature_conf: Dict[str, Dict] = None,
        estimator_params=tuple(),
        fallback=False,
    ):
        self.group_feature = group_feature
        self.model_conf = model_conf
        self.feature_conf = feature_conf
        self.estimator_params = estimator_params
        self.fallback = fallback
        self.components_ = ModelStructure.from_config(
            model_conf, feature_conf
        ).components
        self.estimators_ = OrderedDict({})
        self.added_features_ = []
        self.encoders_ = {
            "main_effects": defaultdict(dict),
            "interactions": defaultdict(dict),
        }

    @property
    def n_parameters(self):
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError as exc:
            raise ValueError(
                "The number of parameters is acceccible only after "
                "the model has been fitted"
            ) from exc
        else:
            n_parameters = 0
            for name, est in self.estimators_.items():
                if name != "_global_":
                    n_parameters += est.n_parameters
            return n_parameters

    @property
    def dof(self):
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError as exc:
            raise ValueError(
                "The degrees of freedom are acceccible only after "
                "the model has been fitted"
            ) from exc
        else:
            dof = 0
            for name, est in self.estimators_.items():
                if name != "_global_":
                    dof += est.dof
            return dof

    def _fit_single_group(self, group_name, model_structure, X, y):
        try:
            params = (
                dict(self.estimator_params) if self.estimator_params is not None else {}
            )

            estimator = LinearPredictor(model_structure=model_structure, **params)
            estimator = estimator.fit(X, y)
        except Exception as e:
            raise type(e)(f"Exception for group {group_name}: {e}")
        else:
            return estimator

    def _predict_single_group(self, group_name, X, include_components):
        """Predict a single group by getting its estimator"""
        try:
            estimator = self.estimators_[group_name]
        except KeyError:
            if self.fallback:
                estimator = self.estimators_["_global_"]
            else:
                raise ValueError(f"Found new group {group_name} during predict")
        finally:
            pred = estimator.predict(X, include_components=include_components)
            if not isinstance(pred, (pd.Series, pd.DataFrame)):
                pred = pd.DataFrame(
                    data=pred, index=X.index, columns=[self.target_name_]
                )
            return pred

    def _update_local_conf(self, conf, X, y=None, fitting=True):
        for name, props in conf["main_effects"].items():
            if props["type"] == "categorical":
                if fitting:
                    stratify_by = (
                        None
                        if not props["stratify_by"]
                        else [self.group_feature] + props["stratify_by"]
                    )
                    enc = CategoricalEncoder(
                        **dict(
                            {key: val for key, val in props.items() if key != "type"},
                            encode_as="ordinal",
                            stratify_by=stratify_by,
                        )
                    )
                    encoded = enc.fit_transform(X, y).squeeze()
                    self.encoders_["main_effects"][name] = enc
                else:
                    enc = self.encoders_["main_effects"][name]
                    encoded = enc.transform(X).squeeze()

                new_name = "__for__".join((props.get("feature"), name))
                X[new_name] = encoded
                props.update(
                    {
                        "feature": new_name,
                        "max_n_categories": None,
                        "stratify_by": None,
                    }
                )

        for pair_name, pair_props in conf["interactions"].items():
            for name in pair_name:
                props = pair_props[name]
                if props["type"] == "categorical":
                    if fitting:
                        stratify_by = (
                            None
                            if not props["stratify_by"]
                            else [self.group_feature] + props["stratify_by"]
                        )
                        enc = CategoricalEncoder(
                            **dict(
                                {
                                    key: val
                                    for key, val in props.items()
                                    if key != "type"
                                },
                                encode_as="ordinal",
                                stratify_by=stratify_by,
                            )
                        )
                        encoded = enc.fit_transform(X, y).squeeze()
                        self.encoders_["interactions"][pair_name].update({name: enc})
                    else:
                        enc = self.encoders_["interactions"][pair_name][name]
                        encoded = enc.transform(X).squeeze()

                    new_name = "__for__".join(
                        (props.get("feature"), ":".join(pair_name))
                    )
                    X[new_name] = encoded
                    props.update(
                        {
                            "feature": new_name,
                            "max_n_categories": None,
                            "stratify_by": None,
                        }
                    )
        return (conf, X) if fitting else X

    def _create_new_features(self):
        added_features = self.components_.pop("add_features")
        self.components_["add_features"] = {}

        if added_features:
            for _, props in added_features.items():
                fgen_type = props.pop("type")
                if isinstance(fgen_type, str):
                    if self.feature_conf is None:
                        raise ValueError(
                            "A mapping between types and classes has not been provided."
                        )

                    targets = self.feature_conf.get(fgen_type)
                    if targets is None:
                        raise ValueError(
                            f"Type {fgen_type} not found in provided mapping"
                        )

                    class_obj = parse_encoder_definition(targets["generate"])
                    self.added_features_.append(class_obj(**props))
                else:
                    self.added_features_.append(fgen_type)

    def fit(self, X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series]):
        """Fit the estimator with the available data.

        Args:
            X (pandas.DataFrame): Input data.
            y (pandas.Series or pandas.DataFrame): Target data.

        Raises:
            Exception: If the estimator is re-fitted. An estimator object can only be
                fitted once.
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If the target data does not pass the checks of `utils.check_y`.

        Returns:
            GroupedPredictor: Fitted estimator.
        """
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError:
            pass
        else:
            raise Exception(
                "Estimator object can only be fit once. Instantiate a new object."
            )

        self._create_new_features()
        # Apply the feature generators
        if self.added_features_:
            X = reduce(
                lambda _df, trans: trans.fit_transform(_df), self.added_features_, X
            )

        X = check_X(X, exists=self.group_feature)
        if self.fallback and ("_global_" in X[self.group_feature]):
            raise ValueError(
                "Name `_global_` is reserved and cannot be used as a group name"
            )
        y = check_y(y, index=X.index)
        self.target_name_ = y.columns[0]

        local_model_conf = copy.deepcopy(self.components_)
        local_model_conf, X = self._update_local_conf(
            local_model_conf, X, y=y, fitting=True
        )

        for group_name, group_data in X.groupby(self.group_feature):
            self.estimators_[group_name] = self._fit_single_group(
                group_name=group_name,
                model_structure=ModelStructure(local_model_conf, self.feature_conf),
                X=group_data.drop(self.group_feature, axis=1),
                y=y.loc[group_data.index],
            )

        if self.fallback:
            self.estimators_["_global_"] = self._fit_single_group(
                group_name="_global_",
                model_structure=ModelStructure(self.components_, self.feature_conf),
                X=X.drop(self.group_feature, axis=1),
                y=y,
            )

        self.groups_ = as_list(self.estimators_.keys())
        self.fitted_ = True
        return self

    def predict(
        self, X: pd.DataFrame, include_clusters=False, include_components=False
    ):
        """Predict given new input data.

        Args:
            X (pandas.DataFrame): Input data.
            include_clusters (bool, optional): Whether to include the added clusters in the
                returned prediction. Defaults to False.
            include_components (bool, optional): Whether to include the contribution of the
                individual components of the model structure in the returned prediction.
                Defaults to False.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            pandas.DataFrame: The predicted values.
        """
        check_is_fitted(self, "fitted_")
        if self.added_features_:
            X = reduce(
                lambda _df, trans: trans.fit_transform(_df), self.added_features_, X
            )
        X = check_X(X, exists=self.group_feature)

        local_model_conf = copy.deepcopy(self.components_)
        X = self._update_local_conf(local_model_conf, X, fitting=False)

        pred = None
        for group_name, group_data in X.groupby(self.group_feature):
            group_pred = self._predict_single_group(
                group_name=group_name,
                X=group_data.drop(self.group_feature, axis=1),
                include_components=include_components,
            )
            if include_clusters:
                group_pred = pd.concat(
                    (group_pred, group_data[[self.group_feature]]),
                    axis=1,
                    ignore_index=False,
                )
            pred = pd.concat((pred, group_pred), axis=0, ignore_index=False)

        pred = pred.reindex(X.index).dropna()
        return pred



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import logging
import warnings

import numpy as np
import pandas as pd
from pandas.api.types import is_bool_dtype as is_bool
from pandas.api.types import is_categorical_dtype as is_category
from pandas.api.types import is_integer_dtype as is_integer
from pandas.api.types import is_object_dtype as is_object
from scipy.stats import skew, wasserstein_distance
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.cluster import KMeans
from sklearn.compose import ColumnTransformer
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import (
    FunctionTransformer,
    OneHotEncoder,
    OrdinalEncoder,
    SplineTransformer,
    StandardScaler,
)
from sklearn.tree import DecisionTreeRegressor
from sklearn.utils.validation import check_is_fitted

from ..utils import add_constant, as_list, check_X, check_y, maybe_reshape_2d

logger = logging.getLogger("feature-encoding")

UNKNOWN_VALUE = -1


#####################################################################################
# Encode features
#
# All encoders generate numpy arrays
#####################################################################################


class IdentityEncoder(TransformerMixin, BaseEstimator):
    """Create an encoder that returns what it is fed.

    This encoder can act as a linear feature encoder.

    Args:
        feature (str or list of str, optional): The name(s) of the input dataframe's
            column(s) to return. If None, the whole input dataframe will be returned.
            Defaults to None.
        as_filter (bool, optional): If True, the encoder will return all feature labels
            for which "feature in label == True". Defaults to False.
        include_bias (bool, optional): If True, a column of ones is added to the output.
            Defaults to False.

    Raises:
        ValueError: If `as_filter` is True, `feature` cannot include multiple feature names.
    """

    def __init__(self, feature=None, as_filter=False, include_bias=False):
        if as_filter and isinstance(feature, list):
            raise ValueError(
                "If `as_filter` is True, `feature` cannot include multiple feature names"
            )

        self.feature = feature
        self.as_filter = as_filter
        self.include_bias = include_bias
        self.features_ = as_list(feature)

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored.
                Defaults to None.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            IdentityEncoder: Fitted encoder.
        """
        X = check_X(X)

        if self.feature is None:
            n_features_out_ = X.shape[1]
        elif (self.feature is not None) and not self.as_filter:
            n_features_out_ = len(self.features_)
        else:
            n_features_out_ = X.filter(like=self.feature, axis=1).shape[1]

        self.n_features_out_ = int(self.include_bias) + n_features_out_
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If `include_bias` is True and a column with constant values
                already exists in the returned columns.

        Returns:
            numpy array of shape: The selected column subset as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X)

        if (self.feature is not None) and not self.as_filter:
            X = X[self.features_]
        elif self.feature is not None:
            X = X.filter(like=self.feature, axis=1)

        if self.include_bias:
            X = add_constant(X, has_constant="raise")

        return np.array(X)


class SafeOrdinalEncoder(TransformerMixin, BaseEstimator):
    """Encode categorical features as an integer array.

    The encoder converts the features into ordinal integers. This results
    in a single column of integers (0 to n_categories - 1) per feature.

    Args:
        feature (str or list of str, optional): The names of the columns to
            encode. If None, all categorical columns will be encoded. Defaults
            to None.
        unknown_value (int, optional): This parameter will set the encoded value
            for unknown categories. It has to be distinct from the values used to
            encode any of the categories in `fit`. If None, the value `-1` is used.
            During `transform`, unknown categories will be replaced using the most
            frequent value along each column. Defaults to None.
    """

    def __init__(self, feature=None, unknown_value=None):
        self.feature = feature
        self.unknown_value = unknown_value
        self.features_ = as_list(feature)

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            SafeOrdinalEncoder: Fitted encoder.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
        """
        X, categorical_cols, _ = check_X(X, exists=self.features_, return_col_info=True)

        if not self.features_:
            self.features_ = categorical_cols
        else:
            for name in self.features_:
                if pd.api.types.is_float_dtype(X[name]):
                    raise ValueError("The encoder is applied on numerical data")

        self.feature_pipeline_ = Pipeline(
            [
                (
                    "select",
                    ColumnTransformer(
                        [("select", "passthrough", self.features_)], remainder="drop"
                    ),
                ),
                (
                    "encode_ordinal",
                    OrdinalEncoder(
                        handle_unknown="use_encoded_value",
                        unknown_value=self.unknown_value or UNKNOWN_VALUE,
                        dtype=np.int16,
                    ),
                ),
                (
                    "impute_unknown",
                    SimpleImputer(
                        missing_values=self.unknown_value or UNKNOWN_VALUE,
                        strategy="most_frequent",
                    ),
                ),
            ]
        )
        # Fit the pipeline
        self.feature_pipeline_.fit(X)
        self.n_features_out_ = len(self.features_)
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            numpy array of shape: The encoded column subset as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X, exists=self.features_)
        return self.feature_pipeline_.transform(X)


class SafeOneHotEncoder(TransformerMixin, BaseEstimator):
    """Encode categorical features in a one-hot form.

    The encoder uses a `SafeOrdinalEncoder`to first encode the feature as an
    integer array and then a `sklearn.preprocessing.OneHotEncoder` to encode
    the features as an one-hot array.

    Args:
        feature (str or list of str, optional): The names of the columns to
            encode. If None, all categorical columns will be encoded. Defaults
            to None.
        unknown_value (int, optional): This parameter will set the encoded value
            of unknown categories. It has to be distinct from the values used to
            encode any of the categories in `fit`. If None, the value `-1` is used.
            During `transform`, unknown categories will be replaced using the most
            frequent value along each column. Defaults to None.
    """

    def __init__(self, feature=None, unknown_value=None):
        self.feature = feature
        self.unknown_value = unknown_value
        self.features_ = as_list(feature)

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            SafeOneHotEncoder: Fitted encoder.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If the encoder is applied on numerical (float) data.
        """
        X, categorical_cols, _ = check_X(X, exists=self.features_, return_col_info=True)

        if not self.features_:
            self.features_ = categorical_cols
        else:
            for name in self.features_:
                if pd.api.types.is_float_dtype(X[name]):
                    raise ValueError("The encoder is applied on numerical data")

        self.feature_pipeline_ = Pipeline(
            [
                (
                    "encode_ordinal",
                    SafeOrdinalEncoder(
                        feature=self.features_,
                        unknown_value=self.unknown_value or UNKNOWN_VALUE,
                    ),
                ),
                ("one_hot", OneHotEncoder(drop=None, sparse=False)),
            ]
        )
        # Fit the pipeline
        self.feature_pipeline_.fit(X)

        self.n_features_out_ = 0
        for category in self.feature_pipeline_["one_hot"].categories_:
            self.n_features_out_ += len(category)

        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            numpy array of shape: The encoded column subset as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X, exists=self.features_)
        return self.feature_pipeline_.transform(X)


class TargetClusterEncoder(TransformerMixin, BaseEstimator):
    """Encode a categorical feature as clusters of the target's values.

    The purpose of this encoder is to reduce the cardinality of a categorical
    feature. This encoder does not replace unknown values with the most frequent
    one during `transform`. It just assigns them the value of `unknown_value`.

    Args:
        feature (str): The name of the categorical feature to transform. This
            encoder operates on a single feature.
        max_n_categories (int, optional): The maximum number of categories to
            produce. Defaults to None.
        stratify_by (str or list of str, optional): If not None, the encoder
            will first stratify the categorical feature into groups that have
            similar values of the features in `stratify_by`, and then cluster
            based on the relationship between the categorical feature and the
            target. It is used only if the number of unique categories minus
            the `excluded_categories` is larger than `max_n_categories`.
            Defaults to None.
        excluded_categories (str or list of str, optional): The names of the
            categories to be excluded from the clustering process. These categories
            will stay intact by the encoding process, so they cannot have the
            same values as the encoder's results (the encoder acts as an
            ``OrdinalEncoder`` in the sense that the feature is converted into
            a column of integers 0 to n_categories - 1). Defaults to None.
        unknown_value (int, optional): This parameter will set the encoded value of
            unknown categories. It has to be distinct from the values used to encode
            any of the categories in `fit`. If None, the value `-1` is used. Defaults
            to None.
        min_samples_leaf (int, optional): The minimum number of samples required to be
            at a leaf node of the decision tree model that is used for stratifying the
            categorical feature if `stratify_by` is not None. The actual number that will
            be passed to the tree model is `min_samples_leaf` multiplied by the number of
            unique values in the categorical feature to transform. Defaults to 1.
        max_features (int, float or {"auto", "sqrt", "log2"}, optional): The number of
            features that the decision tree considers when looking for the best split:

                - If int, then consider `max_features` features at each split of the decision
                  tree

                - If float, then `max_features` is a fraction and `int(max_features * n_features)`
                  features are considered at each split

                - If "auto", then `max_features=n_features`

                - If "sqrt", then `max_features=sqrt(n_features)`

                - If "log2", then `max_features=log2(n_features)`

                - If None, then `max_features=n_features`

            Defaults to "auto".
        random_state (int or RandomState instance, optional): Controls the randomness of
            the decision tree estimator. To obtain a deterministic behaviour during its
            fitting, ``random_state`` has to be fixed to an integer. Defaults to None.
    """

    def __init__(
        self,
        *,
        feature,
        max_n_categories,
        stratify_by=None,
        excluded_categories=None,
        unknown_value=None,
        min_samples_leaf=5,
        max_features="auto",
        random_state=None,
    ):
        self.feature = feature
        self.max_n_categories = max_n_categories
        self.stratify_by = stratify_by
        self.excluded_categories = excluded_categories
        self.unknown_value = unknown_value
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.random_state = random_state
        self.stratify_by_ = as_list(stratify_by)
        self.excluded_categories_ = as_list(excluded_categories)

    def fit(self, X: pd.DataFrame, y: pd.DataFrame):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (pandas.DataFrame of shape (n_samples, 1)): The target dataframe.

        Returns:
            TargetClusterEncoder: Fitted encoder.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If the encoder is applied on numerical (float) data.
            ValueError: If any of the values in `excluded_categories` is not found in
                the input data.
            ValueError: If the number of categories left after removing all in
                `excluded_categories` is not larger than `max_n_categories`.
        """
        X = check_X(X, exists=[self.feature] + self.stratify_by_)
        if pd.api.types.is_float_dtype(X[self.feature]):
            raise ValueError("The encoder is applied on numerical data")

        y = check_y(y, index=X.index)
        self.target_name_ = y.columns[0]

        X = X.merge(y, left_index=True, right_index=True)

        if self.excluded_categories_:
            unique_vals = X[self.feature].unique()
            for value in self.excluded_categories_:
                if value not in unique_vals:
                    raise ValueError(
                        f"Value {value} of `excluded_categories` not found "
                        f"in the {self.feature} data."
                    )

            mask = X[self.feature].isin(self.excluded_categories_)
            X = X.loc[~mask]
            if len(X) == 0:
                raise ValueError(
                    "No categories left after removing all in `excluded_categories`."
                )
            if X[self.feature].nunique() <= self.max_n_categories:
                raise ValueError(
                    "The number of categories left after removing all in `excluded_categories` "
                    "must be larger than `max_n_categories`."
                )

        if not self.stratify_by_:
            self.mapping_ = self._cluster_without_stratify(X)
        else:
            self.mapping_ = self._cluster_with_stratify(X)

        if self.excluded_categories_:
            for i, cat in enumerate(self.excluded_categories_):
                self.mapping_.update({cat: self.max_n_categories + i})

        self.n_features_out_ = 1
        self.fitted_ = True
        return self

    def _cluster_without_stratify(self, X):
        reference = np.array(X[self.target_name_])
        X = X.groupby(self.feature)[self.target_name_].agg(
            ["mean", "std", skew, lambda x: wasserstein_distance(x, reference)]
        )
        X.fillna(value=1, inplace=True)

        X_to_cluster = StandardScaler().fit_transform(X)
        n_clusters = min(X_to_cluster.shape[0], self.max_n_categories)
        clusterer = KMeans(n_clusters=n_clusters)

        with warnings.catch_warnings(record=True) as warning:
            cluster_labels = pd.Series(
                data=clusterer.fit_predict(X_to_cluster), index=X.index
            )
            for w in warning:
                logger.warning(str(w))
        return cluster_labels.to_dict()

    def _cluster_with_stratify(self, X):
        X_train = None
        for col in self.stratify_by_:
            if (
                is_bool(X[col])
                or is_object(X[col])
                or is_category(X[col])
                or is_integer(X[col])
            ):
                X_train = pd.concat((X_train, pd.get_dummies(X[col])), axis=1)
                X_train.columns = X_train.columns.astype(str)
            else:
                X_train = pd.concat((X_train, X[col]), axis=1)

        y_train = X[self.target_name_]
        n_categories = X[self.feature].nunique()

        min_samples_leaf = n_categories * int(self.min_samples_leaf)
        model = DecisionTreeRegressor(
            min_samples_leaf=min_samples_leaf,
            max_features=self.max_features,
            random_state=self.random_state,
        )
        model = model.fit(X_train, y_train)
        leaf_ids = model.apply(X_train)
        uniq_ids = np.unique(leaf_ids)
        leaf_samples = [np.where(leaf_ids == id)[0] for id in uniq_ids]

        X_to_cluster = pd.DataFrame(
            index=X[self.feature].unique(), columns=range(len(leaf_samples))
        )
        for i, idx in enumerate(leaf_samples):
            subset = X.iloc[idx][[self.feature, self.target_name_]]
            a = subset.groupby(self.feature)[self.target_name_].mean()
            a = a.reindex(X_to_cluster.index)
            X_to_cluster.iloc[:, i] = a

        X_to_cluster = X_to_cluster.fillna(X_to_cluster.median())
        n_clusters = min(X_to_cluster.shape[0], self.max_n_categories)

        clusterer = KMeans(n_clusters=n_clusters)
        with warnings.catch_warnings(record=True) as warning:
            cluster_labels = pd.Series(
                data=clusterer.fit_predict(X_to_cluster), index=X_to_cluster.index
            )
            for w in warning:
                logger.warning(str(w))
        return cluster_labels.to_dict()

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.

        Returns:
            numpy array: The encoded column subset as a numpy array.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X, exists=self.feature)

        return maybe_reshape_2d(
            np.array(
                X[self.feature].map(
                    lambda x: int(
                        self.mapping_.get(x, self.unknown_value or UNKNOWN_VALUE)
                    )
                )
            )
        )


class CategoricalEncoder(TransformerMixin, BaseEstimator):
    """Encode categorical features.

    If `max_n_categories` is not `None` and the number of unique values of the
    categorical feature is larger than the `max_n_categories` minus the
    `excluded_categories`, the `TargetClusterEncoder` will be called.

    If `encode_as = 'onehot'`, the result comes from a `TargetClusterEncoder` +
    `SafeOneHotEncoder` pipeline, otherwise from a `TargetClusterEncoder` +
    `SafeOrdinalEncoder` one.

    Args:
        feature (str): The name of the categorical feature to transform. This
            encoder operates on a single feature.
        max_n_categories (int, optional): The maximum number of categories to
            produce. Defaults to None.
        stratify_by (str or list of str, optional): If not None, the encoder
            will first stratify the categorical feature into groups that have
            similar values of the features in `stratify_by`, and then cluster
            based on the relationship between the categorical feature and the
            target. It is used only if the number of unique categories minus
            the `excluded_categories` is larger than `max_n_categories`.
            Defaults to None.
        excluded_categories (str or list of str, optional): The names of the
            categories to be excluded from the clustering process. These categories
            will stay intact by the encoding process, so they cannot have the
            same values as the encoder's results (the encoder acts as an
            ``OrdinalEncoder`` in the sense that the feature is converted into
            a column of integers 0 to n_categories - 1). Defaults to None.
        unknown_value (int, optional): This parameter will set the encoded value of
            unknown categories. It has to be distinct from the values used to encode
            any of the categories in `fit`. If None, the value `-1` is used. Defaults
            to None.
        min_samples_leaf (int, optional): The minimum number of samples required to be
            at a leaf node of the decision tree model that is used for stratifying the
            categorical feature if `stratify_by` is not None. The actual number that will
            be passed to the tree model is `min_samples_leaf` multiplied by the number of
            unique values in the categorical feature to transform. Defaults to 1.
        max_features (int, float or {"auto", "sqrt", "log2"}, optional): The number of
            features that the decision tree considers when looking for the best split:

                - If int, then consider `max_features` features at each split of the decision
                  tree

                - If float, then `max_features` is a fraction and `int(max_features * n_features)`
                  features are considered at each split

                - If "auto", then `max_features=n_features`

                - If "sqrt", then `max_features=sqrt(n_features)`

                - If "log2", then `max_features=log2(n_features)`

                - If None, then `max_features=n_features`

            Defaults to "auto".
        random_state (int or RandomState instance, optional): Controls the randomness of
            the decision tree estimator. To obtain a deterministic behaviour during its
            fitting, ``random_state`` has to be fixed to an integer. Defaults to None.
        encode_as ({'onehot', 'ordinal'}, optional): Method used to encode the transformed
            result.

                - If "onehot", encode the transformed result with one-hot encoding and return a
                  dense array

                - If "ordinal", encode the transformed result as integer values

            Defaults to "onehot".
    """

    def __init__(
        self,
        *,
        feature,
        max_n_categories=None,
        stratify_by=None,
        excluded_categories=None,
        unknown_value=None,
        min_samples_leaf=1,
        max_features="auto",
        random_state=None,
        encode_as="onehot",
    ):
        self.feature = feature
        self.max_n_categories = max_n_categories
        self.stratify_by = stratify_by
        self.excluded_categories = excluded_categories
        self.unknown_value = unknown_value
        self.min_samples_leaf = min_samples_leaf
        self.max_features = max_features
        self.random_state = random_state
        self.encode_as = encode_as
        self.excluded_categories_ = as_list(excluded_categories)

    def _to_pandas(self, arr: np.ndarray):
        return pd.DataFrame(arr, columns=[self.feature])

    def fit(self, X: pd.DataFrame, y: pd.DataFrame = None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (pandas.DataFrame of shape (n_samples, 1), optional): The target dataframe.
                Defaults to None.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If the encoder is applied on numerical (float) data.
            ValueError: If the number of categories minus the `excluded_categories`
                is larger than `max_n_categories` but target values (y) are not
                provided.
            ValueError: If any of the values in `excluded_categories` is not found in
                the input data.

        Returns:
            CategoricalEncoder: Fitted encoder.
        """
        X = check_X(X, exists=self.feature)
        if pd.api.types.is_float_dtype(X[self.feature]):
            raise ValueError("The encoder is applied on numerical data")

        n_categories = X[self.feature].nunique()
        use_target = (self.max_n_categories is not None) and (
            n_categories - len(self.excluded_categories_) > self.max_n_categories
        )

        if use_target and (y is None):
            raise ValueError(
                f"The number of categories to encode: {n_categories - len(self.excluded_categories_)}"
                f" is larger than `max_n_categories`: {self.max_n_categories}. In this case, "
                "the target values must be provided for target-based encoding."
            )

        if not use_target:
            self.feature_pipeline_ = Pipeline(
                [
                    (
                        "encode_features",
                        SafeOneHotEncoder(
                            feature=self.feature, unknown_value=self.unknown_value
                        ),
                    )
                    if self.encode_as == "onehot"
                    else (
                        "encode_features",
                        SafeOrdinalEncoder(
                            feature=self.feature, unknown_value=self.unknown_value
                        ),
                    )
                ]
            )
        else:
            self.feature_pipeline_ = Pipeline(
                [
                    (
                        "reduce_dimension",
                        TargetClusterEncoder(
                            feature=self.feature,
                            stratify_by=self.stratify_by,
                            max_n_categories=self.max_n_categories,
                            excluded_categories=self.excluded_categories,
                            unknown_value=self.unknown_value,
                            min_samples_leaf=self.min_samples_leaf,
                            max_features=self.max_features,
                            random_state=self.random_state,
                        ),
                    ),
                    (
                        "to_pandas",
                        FunctionTransformer(self._to_pandas),
                    ),
                    (
                        "encode_features",
                        SafeOneHotEncoder(
                            feature=self.feature, unknown_value=self.unknown_value
                        ),
                    )
                    if self.encode_as == "onehot"
                    else (
                        "encode_features",
                        SafeOrdinalEncoder(
                            feature=self.feature, unknown_value=self.unknown_value
                        ),
                    ),
                ]
            )

        # Fit the pipeline
        self.feature_pipeline_.fit(X, y)
        self.n_features_out_ = self.feature_pipeline_["encode_features"].n_features_out_
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            numpy array: The encoded features as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X, exists=self.feature)
        return self.feature_pipeline_.transform(X)


class SplineEncoder(TransformerMixin, BaseEstimator):
    """Generate univariate B-spline bases for features.

    The encoder generates a matrix consisting of `n_splines=n_knots + degree - 1`
    spline basis functions (B-splines) of polynomial order=`degree` for the given
    feature.

    Args:
        feature (str): The name of the column to encode.
        n_knots (int, optional): Number of knots of the splines if `knots` equals one
            of {'uniform', 'quantile'}. Must be larger or equal 2. Ignored if `knots`
            is array-like. Defaults to 5.
        degree (int, optional): The polynomial degree of the spline basis. Must be a
            non-negative integer. Defaults to 3.
        strategy ({'uniform', 'quantile'} or array-like of shape (n_knots, n_features),
            optional): Set knot positions such that first knot <= features <= last knot.

                - If 'uniform', `n_knots` number of knots are distributed uniformly
                  from min to max values of the features (each bin has the same width)

                - If 'quantile', they are distributed uniformly along the quantiles of
                  the features (each bin has the same number of observations)

                - If an array-like is given, it directly specifies the sorted knot
                  positions including the boundary knots. Note that, internally,
                  `degree` number of knots are added before the first knot, the same
                  after the last knot

            Defaults to "uniform".
        extrapolation ({'error', 'constant', 'linear', 'continue'}, optional): If 'error',
            values outside the min and max values of the training features raises a `ValueError`.
            If 'constant', the value of the splines at minimum and maximum value of the features
            is used as constant extrapolation. If 'linear', a linear extrapolation is used. If
            'continue', the splines are extrapolated as is, option `extrapolate=True` in
            `scipy.interpolate.BSpline`.
            Defaults to "constant".
        include_bias (bool, optional): If False, then the last spline element inside the data
            range of a feature is dropped. As B-splines sum to one over the spline basis functions
            for each data point, they implicitly include a bias term. Defaults to True.
        order ({'C', 'F'}, optional): Order of output array. 'F' order is faster to compute, but
            may slow down subsequent estimators. Defaults to "C".
    """

    def __init__(
        self,
        *,
        feature,
        n_knots=5,
        degree=3,
        strategy="uniform",
        extrapolation="constant",
        include_bias=True,
        order="C",
    ):
        self.feature = feature
        self.n_knots = n_knots
        self.degree = degree
        self.strategy = strategy
        self.extrapolation = extrapolation
        self.include_bias = include_bias
        self.order = order

    def fit(self, X: pd.DataFrame, y=None, sample_weight=None):
        """Fit the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The data to fit.
            y (None, optional): Ignored. Defaults to None.
            sample_weight (array-like of shape (n_samples,), optional): Individual
                weights for each sample. Used to calculate quantiles if `strategy="quantile"`.
                For `strategy="uniform"`, zero weighted observations are ignored for finding
                the min and max of `X`. Defaults to None.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            SplineEncoder: Fitted encoder.
        """
        X = check_X(X, exists=self.feature)
        self.encoder_ = SplineTransformer(
            n_knots=self.n_knots,
            degree=self.degree,
            knots=self.strategy,
            extrapolation=self.extrapolation,
            include_bias=self.include_bias,
            order=self.order,
        )

        self.encoder_.fit(X[[self.feature]])
        self.n_features_out_ = self.encoder_.n_features_out_
        self.fitted_ = True
        return self

    def transform(self, X):
        """Transform the feature data to B-splines.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The data to transform.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.

        Returns:
            numpy.ndarray: The B-splines matrix.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X, exists=self.feature)
        return self.encoder_.transform(X[[self.feature]])



# -*- coding: utf-8 -*-

import importlib.resources

import feature_encoders.config

with importlib.resources.path(feature_encoders.config, "__init__.py") as init_path:
    CONF_PATH = init_path.resolve().parent



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import importlib

_DEFAULT_PACKAGES = ["feature_encoders.", ""]


def load_obj(obj_path: str, default_obj_path: str = ""):
    """Extract an object from a given path.

    From https://github.com/quantumblacklabs/kedro/blob/0.17.5/kedro/utils.py

    Args:
        obj_path: Path to an object to be extracted, including the object name.
        default_obj_path: Default object path.

    Returns:
        Extracted object.

    Raises:
        AttributeError: When the object does not have the given named attribute.
    """
    obj_path_list = obj_path.rsplit(".", 1)
    obj_path = obj_path_list.pop(0) if len(obj_path_list) > 1 else default_obj_path
    obj_name = obj_path_list[0]
    try:
        module_obj = importlib.import_module(obj_path)
    except ModuleNotFoundError:
        return None

    if not hasattr(module_obj, obj_name):
        return None
    return getattr(module_obj, obj_name)


def parse_encoder_definition(class_obj: str):
    """Parse and instantiate an encoder class using the configuration provided.

    Args:
        config: Encoder config dictionary. It *must* contain the `type` key
            with fully qualified class name.

    Raises:
        ValueError: If the function fails to parse the configuration provided.

    Returns:
        2-tuple: (Encoder class object, configuration dictionary)
    """
    if isinstance(class_obj, str):
        if len(class_obj.strip(".")) != len(class_obj):
            raise ValueError(
                "`type` class path does not support relative "
                "paths or paths ending with a dot."
            )

        class_paths = [prefix + class_obj for prefix in _DEFAULT_PACKAGES]

        trials = [load_obj(class_path) for class_path in class_paths]
        try:
            class_obj = next(obj for obj in trials if obj is not None)
        except StopIteration as exc:
            raise ValueError(
                f"Class `{class_obj}` not found or one of its dependencies"
                f"has not been installed."
            ) from exc

    return class_obj



from ._encoders import (
    CategoricalEncoder,
    IdentityEncoder,
    SafeOneHotEncoder,
    SafeOrdinalEncoder,
    SplineEncoder,
    TargetClusterEncoder,
)
from ._interactions import (
    ICatEncoder,
    ICatLinearEncoder,
    ICatSplineEncoder,
    ISplineEncoder,
    ProductEncoder,
)

__all__ = [
    "CategoricalEncoder",
    "IdentityEncoder",
    "ICatEncoder",
    "ICatLinearEncoder",
    "ICatSplineEncoder",
    "ISplineEncoder",
    "ProductEncoder",
    "SafeOneHotEncoder",
    "SafeOrdinalEncoder",
    "SplineEncoder",
    "TargetClusterEncoder",
]



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import glob
from typing import Any, Union

import numpy as np
import pandas as pd
import scipy
from omegaconf import OmegaConf
from pandas.api.types import is_bool_dtype as is_bool
from pandas.api.types import is_categorical_dtype as is_category
from pandas.api.types import is_datetime64_any_dtype as is_datetime
from pandas.api.types import is_integer_dtype as is_integer
from pandas.api.types import is_object_dtype as is_object
from sklearn.utils import check_array
from sklearn.utils.validation import column_or_1d

from feature_encoders.settings import CONF_PATH


def maybe_reshape_2d(arr: np.ndarray):
    """Reshape an array (if needed) so it's always 2-d and long.

    Args:
        arr (numpy.ndarray): The input array.

    Returns:
        numpy.ndarray: The reshaped array.
    """
    if arr.ndim < 2:
        arr = arr.reshape(-1, 1)
    return arr


def as_list(val: Any):
    """Cast input as list.

    Helper function, always returns a list of the input value.
    """
    if isinstance(val, str):
        return [val]
    if hasattr(val, "__iter__"):
        return list(val)
    if val is None:
        return []
    return [val]


def as_series(x: Union[np.ndarray, pd.Series, pd.DataFrame]):
    """Cast an iterable to a Pandas Series object."""
    if isinstance(x, pd.Series):
        return x
    if isinstance(x, pd.DataFrame):
        return x.iloc[:, 0]
    else:
        return pd.Series(column_or_1d(x))


def get_categorical_cols(X: pd.DataFrame, int_is_categorical=True):
    """Return the names of the categorical columns in the input DataFrame.

    Args:
        X (pandas.DataFrame): Input dataframe.
        int_is_categorical (bool, optional): If True, integer types are
            considered categorical. Defaults to True.

    Returns:
        list: The names of categorical columns in the input DataFrame.
    """
    obj_cols = []
    for col in X.columns:
        # check if it is date
        if is_datetime(X[col]):
            continue
        # check if it is bool, object or category
        if is_bool(X[col]) or is_object(X[col]) or is_category(X[col]):
            obj_cols.append(col)
            continue
        # check if it is integer
        if int_is_categorical and is_integer(X[col]):
            obj_cols.append(col)
            continue
    return obj_cols


def get_datetime_data(X: pd.DataFrame, col_name=None):
    """Get datetime information from the input dataframe.

    Args:
        X (pandas.DataFrame): The input dataframe.
        col_name (str, optional): The name of the column that contains
            datetime information. If None, it is assumed that the datetime
            information is provided by the input dataframe's index.
            Defaults to None.

    Returns:
        pandas.Series: The datetime information.
    """
    if col_name is not None:
        dt_column = X[col_name]
    else:
        dt_column = X.index.to_series()

    col_dtype = dt_column.dtype
    if isinstance(col_dtype, pd.core.dtypes.dtypes.DatetimeTZDtype):
        col_dtype = np.datetime64
    if not np.issubdtype(col_dtype, np.datetime64):
        dt_column = pd.to_datetime(dt_column, infer_datetime_format=True)
    return dt_column


def check_X(
    X: pd.DataFrame, exists=None, int_is_categorical=True, return_col_info=False
):
    """Perform a series of checks on the input dataframe.

    Args:
        X (pamdas.DataFrame): The input dataframe.
        exists (str or list of str, optional): Names of columns that must be present
            in the input dataframe. Defaults to None.
        int_is_categorical (bool, optional): If True, integer types are considered
            categorical. Defaults to True.
        return_col_info (bool, optional): If True, the function will return the names
            of the categorical and the names of the numerical columns, in addition to
            the provided dataframe. Defaults to False.

    Raises:
        ValueError: If the input is not a pandas DataFrame.
        ValueError: If any of the column names in `exists` are not found in the input.
        ValueError: If Nan or inf values are found in the provided input data.

    Returns:
        pandas.DataFrame if `return_col_info` is False else (pandas.DataFrame, list, list)
    """
    if not isinstance(X, pd.DataFrame):
        raise ValueError("Input values are expected as pandas DataFrames.")

    exists = as_list(exists)
    for name in exists:
        if name not in X:
            raise ValueError(f"Regressor {name} missing from dataframe")

    categorical_cols = get_categorical_cols(X, int_is_categorical=int_is_categorical)
    numeric_cols = X.columns.difference(categorical_cols)

    if (len(categorical_cols) > 0) and X[categorical_cols].isnull().values.any():
        raise ValueError("Found NaN values in input's categorical data")
    if (len(numeric_cols) > 0) and np.any(~np.isfinite(X[numeric_cols])):
        raise ValueError("Found NaN or Inf values in input's numerical data")

    if return_col_info:
        return X, categorical_cols, numeric_cols
    return X


def check_y(y: Union[pd.Series, pd.DataFrame], index=None):
    """Perform a series of checks on the input dataframe.

    The checks are carried out by `sklearn.utils.check_array`.

    Args:
        y (Union[pandas.Series, pandas.DataFrame]): The input dataframe.
        index (Union[pandas.Index, pandas.DatetimeIndex], optional): An index to compare
            with the input dataframe's index. Defaults to None.

    Raises:
        ValueError: If the input is neither a pandas Series nor a pandas DataFrame with
            only a single column.
        ValueError: If the input data has different index than the one that was provided
            for comparison (if `index` is not None).

    Returns:
        pandas.DataFrame: The validated input data.
    """
    if isinstance(y, pd.DataFrame) and (y.shape[1] == 1):
        target_name = y.columns[0]
    elif isinstance(y, pd.Series):
        target_name = y.name or "_target_values_"
    else:
        raise ValueError(
            "This estimator accepts target inputs as "
            "`pd.Series` or `pd.DataFrame` with only a single column."
        )

    if (index is not None) and not y.index.equals(index):
        raise ValueError(
            "Input data has different index than the one "
            "that was provided for comparison"
        )

    y = pd.DataFrame(
        data=check_array(y, ensure_2d=False), index=y.index, columns=[target_name]
    )
    return y


def tensor_product(a: np.ndarray, b: np.ndarray, reshape=True):
    """Compute the tensor product of two matrices.

    Args:
        a (numpy array of shape (n, m_a)): The first matrix.
        b (numpy array of shape (n, m_b)): The second matrix.
        reshape (bool, optional): Whether to reshape the result to be 2D (n, m_a * m_b)
            or return a 3D tensor (n, m_a, m_b). Defaults to True.

    Raises:
        ValueError: If input arrays are not 2-dimensional.
        ValueError: If both input arrays do not have the same number of samples.

    Returns:
        numpy.ndarray of shape (n, m_a * m_b) if `reshape = True` else of shape (n, m_a, m_b).
    """
    if (a.ndim != 2) or (b.ndim != 2):
        raise ValueError("Inputs must be 2-dimensional")

    na, ma = a.shape
    nb, mb = b.shape

    if na != nb:
        raise ValueError("Both arguments must have the same number of samples")

    if scipy.sparse.issparse(a):
        a = a.A
    if scipy.sparse.issparse(b):
        b = b.A

    product = a[..., :, None] * b[..., None, :]
    if reshape:
        return product.reshape(na, ma * mb)
    return product


def add_constant(
    data: Union[np.ndarray, pd.Series, pd.DataFrame], prepend=True, has_constant="skip"
):
    """Add a column of ones to an array.

    Args:
        data (array-like): A column-ordered design matrix.
        prepend (bool, optional): If true, the constant is in the first column.
            Else the constant is appended (last column). Defaults to True.
        has_constant ({'raise', 'add', 'skip'}, optional): Behavior if ``data``
            already has a constant. The default will return data without adding
            another constant. If 'raise', will raise an error if any column has a
            constant value. Using 'add' will add a column of 1s if a constant column
            is present. Defaults to "skip".

    Returns:
        numpy.ndarray:  The original values with a constant (column of ones).
    """
    x = np.asanyarray(data)
    ndim = x.ndim
    if ndim == 1:
        x = x[:, None]
    elif x.ndim > 2:
        raise ValueError("Only implemented for 2-dimensional arrays")

    is_nonzero_const = np.ptp(x, axis=0) == 0
    is_nonzero_const &= np.all(x != 0.0, axis=0)
    if is_nonzero_const.any():
        if has_constant == "skip":
            return x
        elif has_constant == "raise":
            if ndim == 1:
                raise ValueError("data is constant.")
            else:
                columns = np.arange(x.shape[1])
                cols = ",".join([str(c) for c in columns[is_nonzero_const]])
                raise ValueError(f"Column(s) {cols} are constant.")

    x = [np.ones(x.shape[0]), x]
    x = x if prepend else x[::-1]
    return np.column_stack(x)


def load_config(model="towt", features="default", merge_multiple=False):
    """Load model configuration and feature generator mapping.

    Given `model` and `features`, the function searches for files in:
    ::
        conf_path = str(CONF_PATH)
        model_files = glob.glob(f"{conf_path}/models/{model}.*")
        feature_files = glob.glob(f"{conf_path}/features/{features}.*")

    Args:
        model (str, optional): The name of the model configuration to load.
            Defaults to "towt".
        features (str, optional): The name of the feature generator mapping to
            load. Defaults to "default".
        merge_multiple (bool, optional): If True and more than one files are found when
            searching for either models or features, the contents of the files will ne merged.
            Otherwise, an exception will be raised. Defaults to False.

    Returns:
        (dict, dict): The model configuration and feature mapping as dictionaries.
    """
    conf_path = str(CONF_PATH)

    model_conf = None
    model_files = glob.glob(f"{conf_path}/models/{model}.*")
    if len(model_files) == 0:
        raise ValueError("No model configuration files found")
    elif (len(model_files) > 1) and (not merge_multiple):
        raise ValueError("More than one model configuration files found")
    elif len(model_files) > 1:
        model_conf = OmegaConf.merge(
            *[OmegaConf.load(model_file) for model_file in model_files]
        )
    else:
        model_conf = OmegaConf.load(model_files[0])

    feature_conf = None
    feature_files = glob.glob(f"{conf_path}/features/{features}.*")
    if len(feature_files) == 0:
        raise ValueError("No feature generator mapping files found")
    elif (len(feature_files) > 1) and (not merge_multiple):
        raise ValueError("More than one feature generator mapping files found")
    elif len(feature_files) > 1:
        feature_conf = OmegaConf.merge(
            *[OmegaConf.load(feature_file) for feature_file in feature_files]
        )
    else:
        feature_conf = OmegaConf.load(feature_files[0])

    return OmegaConf.to_container(model_conf), OmegaConf.to_container(feature_conf)



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import copy
from collections import OrderedDict, defaultdict
from functools import reduce
from typing import Dict, Type, Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from ..encode import (
    CategoricalEncoder,
    ICatEncoder,
    ICatLinearEncoder,
    ICatSplineEncoder,
    IdentityEncoder,
    ISplineEncoder,
    ProductEncoder,
    SplineEncoder,
)
from ..validate.schemas import CategoricalSchema, LinearSchema, SplineSchema
from ._parse import parse_encoder_definition

################################################################################
# Utilities
################################################################################


def _validate_feature(props, feature_map):
    fgen_type = props.get("type")
    targets = feature_map.get(fgen_type)
    if targets is None:
        raise ValueError(f"Type {fgen_type} not found in provided mapping")
    class_obj = parse_encoder_definition(targets["validate"])
    validated_props = class_obj(**props).dict()
    return validated_props


def _validate_encoder(props):
    enc_type = props.get("type")
    if enc_type == "categorical":
        validated_props = CategoricalSchema(**props).dict()
    elif enc_type == "linear":
        validated_props = LinearSchema(**props).dict()
    elif enc_type == "spline":
        validated_props = SplineSchema(**props).dict()
    else:
        raise ValueError(f"Type {enc_type} not recognized")
    return validated_props


def _interaction_by_types(left_enc, right_enc):
    left_enc_type = type(left_enc).__name__
    right_enc_type = type(right_enc).__name__

    if (left_enc_type, right_enc_type) == ("CategoricalEncoder", "CategoricalEncoder"):
        return ICatEncoder(left_enc, right_enc)
    elif (left_enc_type, right_enc_type) == ("CategoricalEncoder", "IdentityEncoder"):
        return ICatLinearEncoder(encoder_cat=left_enc, encoder_num=right_enc)
    elif (left_enc_type, right_enc_type) == ("CategoricalEncoder", "SplineEncoder"):
        return ICatSplineEncoder(encoder_cat=left_enc, encoder_num=right_enc)
    elif (left_enc_type, right_enc_type) == ("IdentityEncoder", "IdentityEncoder"):
        return ProductEncoder(left_enc, right_enc)
    elif (left_enc_type, right_enc_type) == ("IdentityEncoder", "CategoricalEncoder"):
        return ICatLinearEncoder(encoder_cat=right_enc, encoder_num=left_enc)
    elif (left_enc_type, right_enc_type) == ("SplineEncoder", "SplineEncoder"):
        return ISplineEncoder(left_enc, right_enc)
    elif (left_enc_type, right_enc_type) == ("SplineEncoder", "CategoricalEncoder"):
        return ICatSplineEncoder(encoder_cat=right_enc, encoder_num=left_enc)
    else:
        raise NotImplementedError(
            f"Interactions between encoder type `{left_enc_type}` "
            f"and encoder type `{right_enc_type}` are not supported"
        )


#######################################################################################
# ModelStructure
#######################################################################################


class ModelStructure:
    """Capture the structure of a linear regression model.

    The class validates and stores the details of a linear regression model: features,
    main effects and interactions.

    Args:
        structure (Dict, optional): A dictionary that includes information about the
            model. Example:
            ::
                {'add_features':
                    {'time':
                        { 'ds': None,
                        'remainder': 'passthrough',
                        'replace': False,
                        'subset': ['month', 'hourofweek']
                        }
                    },
                'main_effects':
                    {'month':
                        { 'feature': 'month',
                        'max_n_categories': None,
                        'encode_as': 'onehot',
                        'interaction_only': False
                        },
                    'tow':
                        { 'feature': 'hourofweek',
                        'max_n_categories': 60,
                        'encode_as': 'onehot',
                        'interaction_only': False
                        },
                    'lin_temperature':
                        { 'feature': 'temperature',
                        'include_bias': False,
                        'interaction_only': False
                        }
                    },
                }
            Defaults to None.
        feature_map (Dict, optional): A mapping between a feature generator name
            and the classes for its validation and creation.
            Example:
            ::
                {'datetime':
                    'validate': 'validate.DatetimeSchema'
                    'generate': 'generate.DatetimeFeatures'
                }
            Defaults to None.
    """

    def __init__(self, structure: Dict = None, feature_map: Dict = None):
        self.structure = structure
        self.feature_map = feature_map
        self.structure_ = (
            {
                "add_features": defaultdict(dict),
                "main_effects": defaultdict(dict),
                "interactions": defaultdict(dict),
            }
            if structure is None
            else structure
        )

    @property
    def components(self):
        return copy.deepcopy(self.structure_)

    @classmethod
    def from_config(cls: Type, config: Dict, feature_map: Dict = None):
        """Create a ModelStructure instance from a configuration file.

        Args:
            config (Dict): A dictionary that includes information about the
                model.
            feature_map (Dict, optional): A mapping between a feature generator
                name and the classes for its validation and creation.
                Defaults to None.

        Returns:
            ModelStructure: A populated ModelStructure instance.
        """
        config = copy.deepcopy(config)
        structure = {
            "add_features": defaultdict(dict),
            "main_effects": defaultdict(dict),
            "interactions": defaultdict(dict),
        }
        temporary = {}

        if "add_features" in config:
            if feature_map is None:
                raise ValueError(
                    "A mapping between feature generator types and classes "
                    "has not been provided."
                )

            for name, props in config["add_features"].items():
                structure["add_features"][name] = _validate_feature(props, feature_map)

        if "regressors" in config:
            for name, props in config["regressors"].items():
                interaction_only = props.pop("interaction_only", False)
                if interaction_only:
                    temporary[name] = props
                else:
                    structure["main_effects"][name] = props

        if "interactions" in config:
            # example of pair_name: temperature, hour
            for pair_name, pair_props in config["interactions"].items():
                pair_name = tuple([x.strip() for x in pair_name.split(",")])
                if len(pair_name) != 2:
                    raise ValueError("Only pairwise interactions are supported.")
                if pair_props is None:
                    pair_props = {}

                for name in pair_name:
                    if name in structure["main_effects"]:
                        props = dict(
                            structure["main_effects"][name],
                            **pair_props.get(name, dict()),
                        )
                        structure["interactions"][pair_name].update({f"{name}": props})

                    elif name in temporary:
                        props = dict(
                            temporary[name],
                            **pair_props.get(name, dict()),
                        )
                        structure["interactions"][pair_name].update({f"{name}": props})

                    elif name in pair_props:
                        structure["interactions"][pair_name].update(
                            {f"{name}": pair_props[name]}
                        )
                    else:
                        raise ValueError(
                            f"The regressor `{name}` has not been added yet and not "
                            "enough information has been provided so that to add it"
                        )
        # validate before store
        for name, props in structure["main_effects"].items():
            structure["main_effects"][name] = _validate_encoder(props)

        for pair_name, pair_props in structure["interactions"].items():
            for name in pair_name:
                structure["interactions"][pair_name][name] = _validate_encoder(
                    pair_props[name]
                )

        return cls(structure=structure, feature_map=feature_map)

    def add_new_feature(
        self, *, name: str, fgen_type: Union[str, BaseEstimator], **kwargs
    ):
        """Add a feature generator.

        Feature generators are applied on the input dataframe with the same order
        that they were added.

        Args:
            name (str): A name for the feature generator.
            fgen_type (str or sklearn-compatible transformer): The feature generator
                to add. If it is a string, the corresponding class will be loaded
                based on the relevant entry in the :attr:`feature_map` dictionary.
            **kwargs: Keyword arguments to be passed during the feature generator
                initialization. Ignored if `fgen` is not a string.

        Raises:
            ValueError: If a feature generator with the same name has already been added.

        Returns:
            ModelStructure: The updated ModelStructure instance.
        """
        if name in self.structure_["add_features"]:
            raise ValueError(f"Feature generator named {name} has already been added")

        self.structure_["add_features"][name].update(
            dict(
                type=fgen_type,
                **kwargs,
            )
        )
        return self

    def add_main_effect(
        self, *, name: str, enc_type: Union[str, BaseEstimator], **kwargs
    ):
        """Add a main effect.

        Args:
            name (str): A name for the main effect.
            enc_type (str or encoder object): The type of the feature encoder to
                apply on the main effect.
            **kwargs: Keyword arguments to be passed during the feature encoder
                initialization. Ignored if `enc_type` is not a string.

        Raises:
            ValueError: If an encoder with the same name has already been added.

        Returns:
            ModelStructure: The updated ModelStructure instance.
        """
        if name in self.structure_["main_effects"]:
            raise ValueError(f"Encoder named {name} has already been added")
        if isinstance(enc_type, str) and (
            enc_type not in ("linear", "spline", "categorical")
        ):
            raise ValueError(f"Encoder type enc_type {enc_type} is not supported")

        self.structure_["main_effects"][name].update(
            dict(
                type=enc_type,
                **kwargs,
            )
        )
        return self

    def add_interaction(
        self,
        *,
        lenc_name: str,
        renc_name: str,
        lenc_type: Union[str, object],
        renc_type: Union[str, object],
        **kwargs,
    ):
        """Add a pairwise interaction.

        Args:
            lenc_name (str): A name for the first part of the interaction pair.
            renc_name (str): A name for the second part of the interaction pair.
            lenc_type (str or encoder object): The type of the feature encoder to
                apply on the first part of the interaction pair.
            renc_type (str or encoder object): The type of the feature encoder to
                apply on the second part of the interaction pair.
            **kwargs: Keyword arguments to be passed during the feature encoders'
                initialization.

        Raises:
            ValueError: If an interaction with the same name `(lenc_name, renc_name)`
                has already been added.

        Returns:
            ModelStructure: The updated ModelStructure instance.

        Example:
        ::
            model = ModelStructure().add_interaction(
                lenc_name="is_Monday",
                renc_name="daily_seasonality",
                lenc_type="categorical",
                renc_type="linear",
                **{
                    is_Monday: {"feature": "is_Monday", "encode_as": "onehot"},
                    daily_seasonality: {"feature": "daily", "as_filter": True},
                },
            )
        """
        if ((lenc_name, renc_name) in self.structure_["interactions"]) or (
            (renc_name, lenc_name) in self.structure_["interactions"]
        ):
            raise ValueError(
                f"Interaction {(lenc_name, renc_name)} has already been added"
            )

        self.structure_["interactions"][(lenc_name, renc_name)][lenc_name] = dict(
            type=lenc_type,
            **kwargs.get(lenc_name, {}),
        )
        self.structure_["interactions"][(lenc_name, renc_name)][renc_name] = dict(
            type=renc_type,
            **kwargs.get(renc_name, {}),
        )
        return self


#######################################################################################
# FeatureComposer
#######################################################################################


class FeatureComposer(TransformerMixin, BaseEstimator):
    """Generate linear features and pairwise interactions.

    Args:
        model_structure (ModelStructure): The structure of a linear regression
            model.
    """

    def __init__(self, model_structure: ModelStructure):
        self.model_structure = model_structure
        self.encoders_ = {
            "main_effects": OrderedDict({}),
            "interactions": OrderedDict({}),
        }
        self.added_features_ = []
        self.train_feature_cols_ = []
        self.component_names_ = []

    def _create_new_features(self):
        for _, props in self.model_structure.components["add_features"].items():
            fgen_type = props.pop("type")
            if isinstance(fgen_type, str):
                if self.model_structure.feature_map is None:
                    raise ValueError(
                        "A mapping between types and classes has not been provided."
                    )

                targets = self.model_structure.feature_map.get(fgen_type)
                if targets is None:
                    raise ValueError(f"Type {fgen_type} not found in provided mapping")

                class_obj = parse_encoder_definition(targets["generate"])
                self.added_features_.append(class_obj(**props))
            else:
                self.added_features_.append(fgen_type)

    def _create_encoders(self):
        for name, props in self.model_structure.components["main_effects"].items():
            enc_type = props.pop("type")
            if isinstance(enc_type, str):
                enc_type = (
                    CategoricalEncoder(**props)
                    if enc_type == "categorical"
                    else SplineEncoder(**props)
                    if enc_type == "spline"
                    else IdentityEncoder(**props)
                )
            self.encoders_["main_effects"][name] = enc_type

        for name, props in self.model_structure.components["interactions"].items():
            left, right = name
            left_enc_type = props[left].pop("type")
            right_enc_type = props[right].pop("type")

            if isinstance(left_enc_type, str):
                left_enc = (
                    CategoricalEncoder(**props[left])
                    if left_enc_type == "categorical"
                    else SplineEncoder(**props[left])
                    if left_enc_type == "spline"
                    else IdentityEncoder(**props[left])
                )
            else:
                left_enc = left_enc_type

            if isinstance(right_enc_type, str):
                right_enc = (
                    CategoricalEncoder(**props[right])
                    if right_enc_type == "categorical"
                    else SplineEncoder(**props[right])
                    if right_enc_type == "spline"
                    else IdentityEncoder(**props[right])
                )
            else:
                right_enc = right_enc_type

            interaction = _interaction_by_types(left_enc, right_enc)
            self.encoders_["interactions"][name] = interaction

    def _main_effects(self, X, y=None, fitting=True):
        for name, encoder in self.encoders_["main_effects"].items():
            if fitting:
                encoder.fit(X, y)
                yield name, encoder.n_features_out_
            else:
                yield name, encoder.transform(X)

    def _interaction_effects(self, X, y=None, fitting=True):
        for name, encoder in self.encoders_["interactions"].items():
            if fitting:
                encoder.fit(X, y)
                yield name, encoder.n_features_out_
            else:
                yield name, encoder.transform(X)

    @property
    def component_matrix(self):
        """Dataframe indicating which columns of the feature matrix correspond
        to which components.

        Returns
        -------
        feature_cols: A binary indicator dataframe. Entry is 1 if that column is used
            in that component.
        """
        if self.train_feature_cols_ is None:
            raise ValueError(
                "The estimator must be fitted before the `component_matrix` can be accessed."
            )

        components = pd.DataFrame(
            {
                "col": np.arange(len(self.train_feature_cols_)),
                "component": [x.split("_delim_")[0] for x in self.train_feature_cols_],
            }
        )
        # Convert to a binary matrix
        feature_cols = pd.crosstab(
            components["col"],
            components["component"],
        ).sort_index(level="col")

        return feature_cols

    def fit(self, X, y=None):
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError:
            pass
        else:
            raise Exception(
                "Estimator object can only be fit once. Instantiate a new object."
            )

        feature_cols = []
        self._create_new_features()
        # Apply the feature generators
        if self.added_features_:
            X = reduce(
                lambda _df, trans: trans.fit_transform(_df), self.added_features_, X
            )

        self._create_encoders()
        # Fit the main effect encoders
        for name, n_features_out_ in self._main_effects(X, y, fitting=True):
            if "_delim_" in name:
                raise ValueError('The name of the regressor cannot include "_delim_"')
            feature_cols.extend([f"{name}_delim_{i}" for i in range(n_features_out_)])

        # Fit the interaction encoders
        for (left, right), n_features_out_ in self._interaction_effects(
            X, y, fitting=True
        ):
            if ("_delim_" in left) or ("_delim_" in right):
                raise ValueError('The name of the regressor cannot include "_delim_"')
            feature_cols.extend(
                [f"{left}:{right}_delim_{i}" for i in range(n_features_out_)]
            )

        self.train_feature_cols_ = feature_cols
        self.component_names_ = self.component_matrix.columns.tolist()
        self.fitted_ = True
        return self

    def transform(self, X):
        check_is_fitted(self, "fitted_")

        if self.added_features_:
            X = reduce(
                lambda _df, trans: trans.fit_transform(_df), self.added_features_, X
            )

        design_matrix = np.zeros((len(X), len(self.train_feature_cols_)))

        # Add the main effects
        for name, features in self._main_effects(X, fitting=False):
            relevant_cols = self.component_matrix.loc[
                self.component_matrix[name] == 1
            ].index
            design_matrix[:, relevant_cols] = features

        # Add the interactions
        for name, features in self._interaction_effects(X, fitting=False):
            left, right = name
            relevant_cols = self.component_matrix.loc[
                self.component_matrix[f"{left}:{right}"] == 1
            ].index
            design_matrix[:, relevant_cols] = features

        return design_matrix



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from typing import Union

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, RegressorMixin
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LinearRegression, Ridge
from sklearn.utils.validation import check_is_fitted

from ..compose import FeatureComposer, ModelStructure
from ..utils import check_X, check_y


class LinearPredictor(RegressorMixin, BaseEstimator):
    """A linear regression model with flexible parameterization.

    Args:
        model_structure (ModelStructure): The structure of a linear regression
            model.
        alpha (float, optional): Regularization strength of the underlying ridge
            regression; must be a positive float. Regularization improves the
            conditioning of the problem and reduces the variance of the estimates.
            Larger values specify stronger regularization. Defaults to 0.01.
        fit_intercept (bool, optional): Whether to fit the intercept for this model.
            If set to false, no intercept will be used in calculations. Defaults to
            False.
    """

    def __init__(
        self, *, model_structure: ModelStructure, alpha=0.01, fit_intercept=False
    ):
        self.model_structure = model_structure
        self.alpha = alpha
        self.fit_intercept = fit_intercept
        self.composer_ = FeatureComposer(model_structure)

    @property
    def n_parameters(self):
        try:
            self.n_parameters_
        except AttributeError as exc:
            raise ValueError(
                "The number of parameters is acceccible only after "
                "the model has been fitted"
            ) from exc
        else:
            return self.n_parameters_

    @property
    def dof(self):
        try:
            self.dof_
        except AttributeError as exc:
            raise ValueError(
                "The degrees of freedom are acceccible only after "
                "the model has been fitted"
            ) from exc
        else:
            return self.dof_

    def fit(self, X: pd.DataFrame, y: Union[pd.DataFrame, pd.Series]):
        """Fit the estimator with the available data.

        Args:
            X (pandas.DataFrame): Input data.
            y (pandas.Series or pandas.DataFrame): Target data.

        Raises:
            Exception: If the estimator is re-fitted. An estimator object can only be
                fitted once.
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If the target data does not pass the checks of `utils.check_y`.

        Returns:
            LinearPredictor: Fitted estimator.
        """
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError:
            pass
        else:
            raise Exception(
                "Estimator object can only be fit once. Instantiate a new object."
            )

        X = check_X(X)
        y = check_y(y, index=X.index)
        self.target_name_ = y.columns[0]

        design_matrix = self.composer_.fit_transform(X, y)
        self.n_parameters_ = design_matrix.shape[1]
        self.dof_ = np.linalg.matrix_rank(design_matrix)

        if self.alpha is None:
            self.base_estimator_ = LinearRegression(fit_intercept=self.fit_intercept)
        else:
            self.base_estimator_ = Ridge(
                alpha=self.alpha, fit_intercept=self.fit_intercept
            )
        self.base_estimator_ = self.base_estimator_.fit(design_matrix, y)
        self.fitted_ = True
        return self

    def predict(self, X: pd.DataFrame, include_components=False):
        """Predict using the given input data.

        Args:
            X (pandas.DataFrame): Input data.
            include_components (bool, optional): If True, the prediction dataframe will
                include also the individual components' contribution to the predicted
                values. Defaults to False.

        Returns:
            pandas.DataFrame: The prediction.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X)
        design_matrix = self.composer_.transform(X)

        prediction = pd.DataFrame(
            data=self.base_estimator_.predict(design_matrix),
            columns=[self.target_name_],
            index=X.index,
        )

        if include_components:
            components = pd.DataFrame(
                0, index=X.index, columns=self.composer_.component_names_
            )
            feature_cols = self.composer_.component_matrix

            for col in components.columns:
                subset = feature_cols[feature_cols[col] == 1].index.to_list()
                coef = self.base_estimator_.coef_.squeeze()
                pred = np.matmul(design_matrix[:, subset], coef[subset])
                components[col] = components[col] + pred

            prediction = pd.concat((prediction, components), axis=1)

        return prediction



from .schemas import (
    CategoricalSchema,
    CyclicalSchema,
    DatetimeSchema,
    LinearSchema,
    SplineSchema,
    TrendSchema,
)

__all__ = [
    "CategoricalSchema",
    "CyclicalSchema",
    "DatetimeSchema",
    "LinearSchema",
    "SplineSchema",
    "TrendSchema",
]



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from collections import OrderedDict
from typing import Union

import pandas as pd
from sklearn.base import BaseEstimator
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import RANSACRegressor, Ridge
from sklearn.utils.validation import check_is_fitted

from ..compose import FeatureComposer, ModelStructure
from ..encode import CategoricalEncoder, IdentityEncoder
from ..generate import CyclicalFeatures, TrendFeatures
from ..utils import as_list, check_X, check_y


class SeasonalPredictor(BaseEstimator):
    """Time series prediction model based on seasonal decomposition.

    Args:
        ds (str, optional): The name of the input dataframe's column that
            contains datetime information. If None, it is assumed that the
            datetime information is provided by the input dataframe's index.
            Defaults to None.
        add_trend (bool, optional): If True, a linear time trend will be
            added. Defaults to False.
        yearly_seasonality (Union[str, bool, int], optional): Fit yearly
            seasonality. Can be 'auto', True, False, or a number of Fourier
            terms to generate. Defaults to "auto".
        weekly_seasonality (Union[str, bool, int], optional): Fit weekly
            seasonality. Can be 'auto', True, False, or a number of Fourier
            terms to generate. Defaults to "auto".
        daily_seasonality (Union[str, bool, int], optional): Fit daily
            seasonality. Can be 'auto', True, False, or a number of Fourier
            terms to generate. Defaults to "auto".
        min_samples (float ([0, 1]), optional): Minimum number of samples
            chosen randomly from original data by the RANSAC (RANdom SAmple
            Consensus) algorithm. Defaults to 0.5.
        alpha (float, optional): Parameter for the underlying ridge estimator
            (`base_estimator`). It must be a positive float. Regularization
            improves the conditioning of the problem and reduces the variance
            of the estimates. Larger values specify stronger regularization.
            Defaults to 0.01.
    """

    def __init__(
        self,
        ds: str = None,
        add_trend: bool = False,
        yearly_seasonality: Union[str, bool, int] = "auto",
        weekly_seasonality: Union[str, bool, int] = "auto",
        daily_seasonality: Union[str, bool, int] = "auto",
        min_samples=0.5,
        alpha=0.01,
    ):
        self.ds = ds
        self.add_trend = add_trend
        self.yearly_seasonality = yearly_seasonality
        self.weekly_seasonality = weekly_seasonality
        self.daily_seasonality = daily_seasonality
        self.min_samples = min_samples
        self.alpha = alpha

        # Set during fitting
        self.seasonalities_ = OrderedDict({})
        self.base_estimator_ = RANSACRegressor(
            base_estimator=Ridge(fit_intercept=True, alpha=alpha),
            min_samples=min_samples,
        )

    def add_seasonality(
        self,
        name: str,
        period: float = None,
        fourier_order: int = None,
        condition_name: str = None,
    ):
        """Add a seasonal component with specified period and number of
        Fourier components.

        If `condition_name is provided`, the input dataframe passed to
        `fit` and `predict` should have a column with the specified
        `condition_name` containing booleans that indicate when to apply
        seasonality.

        Args:
            name (str): The name of the seasonality component.
            period (float, optional): Number of days in one period.
                Defaults to None.
            fourier_order (int, optional): Number of Fourier components
                to use. Defaults to None.
            condition_name (str, optional): The name of the seasonality
                condition. Defaults to None.

        Raises:
            Exception: If the method is called after the estimator is fitted.
            ValueError: If either `period` or `fourier_order` are not provided
                and the seasonality is not in ('daily', 'weekly', 'yearly').

        Returns:
            SeasonalPredictor: The updated estimator object.
        """
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError:
            pass
        else:
            raise Exception("Seasonality must be added prior to model fitting.")

        if name not in ["daily", "weekly", "yearly"]:
            if (period is None) or (fourier_order is None):
                raise ValueError(
                    "When adding custom seasonalities, values for "
                    '"period" and "fourier_order" must be specified.'
                )

        if (period is not None) and (period <= 0):
            raise ValueError("Period must be > 0")
        if (fourier_order is not None) and (fourier_order <= 0):
            raise ValueError("Fourier order must be > 0")

        self.seasonalities_[name] = {
            "period": float(period) if period is not None else None,
            "fourier_order": int(fourier_order) if fourier_order is not None else None,
            "condition_name": condition_name,
        }
        return self

    def _set_seasonalities(self, X):
        dates = X.index.to_series() if self.ds is None else X[self.ds]
        first = dates.min()
        last = dates.max()
        dt = dates.diff()
        time_step = dt.iloc[dt.values.nonzero()[0]].min()

        default_params = {"period": None, "fourier_order": None, "condition_name": None}

        # Set yearly seasonality
        if (self.yearly_seasonality is False) or ("yearly" in self.seasonalities_):
            pass
        elif self.yearly_seasonality is True:
            self.seasonalities_["yearly"] = default_params
        elif self.yearly_seasonality == "auto":
            # Turn on yearly seasonality if there is >=1 years of history
            if last - first >= pd.Timedelta(days=365):
                self.seasonalities_["yearly"] = default_params
        elif self.yearly_seasonality <= 0:
            raise ValueError("Fourier order must be > 0")
        else:
            self.seasonalities_["yearly"] = dict(
                default_params, fourier_order=self.yearly_seasonality
            )

        # Set weekly seasonality
        if (self.weekly_seasonality is False) or ("weekly" in self.seasonalities_):
            pass
        elif self.weekly_seasonality is True:
            self.seasonalities_["weekly"] = default_params
        elif self.weekly_seasonality == "auto":
            # Turn on yearly seasonality if there is >=1 years of history
            if (last - first >= pd.Timedelta(weeks=1)) and (
                time_step < pd.Timedelta(weeks=1)
            ):
                self.seasonalities_["weekly"] = default_params
        elif self.weekly_seasonality <= 0:
            raise ValueError("Fourier order must be > 0")
        else:
            self.seasonalities_["weekly"] = dict(
                default_params, fourier_order=self.weekly_seasonality
            )

        # Set daily seasonality
        if (self.daily_seasonality is False) or ("daily" in self.seasonalities_):
            pass
        elif self.daily_seasonality is True:
            self.seasonalities_["daily"] = default_params
        elif self.daily_seasonality == "auto":
            # Turn on yearly seasonality if there is >=1 years of history
            if (last - first >= pd.Timedelta(days=1)) and (
                time_step < pd.Timedelta(days=1)
            ):
                self.seasonalities_["daily"] = default_params
        elif self.daily_seasonality <= 0:
            raise ValueError("Fourier order must be > 0")
        else:
            self.seasonalities_["daily"] = dict(
                default_params, fourier_order=self.daily_seasonality
            )
        return self

    def _create_composer(self):
        model_structure = ModelStructure()

        if self.add_trend:
            model_structure = model_structure.add_new_feature(
                name="added_trend",
                fgen_type=TrendFeatures(
                    ds=self.ds,
                    name="growth",
                    remainder="passthrough",
                    replace=False,
                ),
            )
            model_structure = model_structure.add_main_effect(
                name="trend",
                enc_type=IdentityEncoder(
                    feature="growth",
                    as_filter=False,
                    include_bias=False,
                ),
            )

        for seasonality, props in self.seasonalities_.items():
            condition_name = props["condition_name"]

            model_structure = model_structure.add_new_feature(
                name=seasonality,
                fgen_type=CyclicalFeatures(
                    seasonality=seasonality,
                    ds=self.ds,
                    period=props.get("period"),
                    fourier_order=props.get("fourier_order"),
                    remainder="passthrough",
                    replace=False,
                ),
            )

            if condition_name is None:
                model_structure = model_structure.add_main_effect(
                    name=seasonality,
                    enc_type=IdentityEncoder(
                        feature=seasonality,
                        as_filter=True,
                        include_bias=False,
                    ),
                )
            else:
                model_structure = model_structure.add_interaction(
                    lenc_name=condition_name,
                    renc_name=seasonality,
                    lenc_type=CategoricalEncoder(
                        feature=condition_name, encode_as="onehot"
                    ),
                    renc_type=IdentityEncoder(
                        feature=seasonality, as_filter=True, include_bias=False
                    ),
                )
        return FeatureComposer(model_structure)

    def _check_input(self, X):
        conditions = [
            props["condition_name"]
            for props in self.seasonalities_.values()
            if props["condition_name"] is not None
        ]

        regressors = as_list(self.ds) + conditions
        X = check_X(X, exists=regressors)

        for condition_name in conditions:
            if not X[condition_name].isin([True, False]).all():
                raise ValueError(f"Found non-boolean in column {condition_name!r}")
        return X

    def fit(self, X: pd.DataFrame, y: pd.DataFrame):
        """Fit the estimator with the available data.

        Args:
            X (pandas.DataFrame): Input data.
            y (pandas.DataFrame): Target data.

        Raises:
            Exception: If the estimator is re-fitted. An estimator object can only be
                fitted once.
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If the target data does not pass the checks of `utils.check_y`.

        Returns:
            SeasonalPredictor: Fitted estimator.
        """
        try:
            check_is_fitted(self, "fitted_")
        except NotFittedError:
            pass
        else:
            raise Exception(
                "Estimator object can only be fit once. Instantiate a new object."
            )

        X = self._check_input(X)
        y = check_y(y, index=X.index)
        self.target_name_ = y.columns[0]

        self._set_seasonalities(X)
        self.composer_ = self._create_composer()

        design_matrix = self.composer_.fit_transform(X, y)
        self.base_estimator_.fit(design_matrix, y)
        self.fitted_ = True
        return self

    def predict(self, X: pd.DataFrame) -> pd.DataFrame:
        """Predict using the given input data.

        Args:
            X (pandas.DataFrame): Input data.

        Returns:
            pandas.DataFrame: The prediction.
        """
        check_is_fitted(self, "fitted_")
        X = self._check_input(X)
        design_matrix = self.composer_.transform(X)

        prediction = pd.DataFrame(
            data=self.base_estimator_.predict(design_matrix),
            index=X.index,
            columns=[self.target_name_],
        )
        return prediction



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

import numbers
from typing import List, Optional, Union

from pydantic import BaseModel, validator


class TrendSchema(BaseModel):
    type: str
    ds: Optional[str] = None
    name: str = "growth"
    remainder: str = "passthrough"
    replace: bool = False

    @validator("remainder")
    def check_remainder(cls, data):
        if data not in ("drop", "passthrough"):
            raise ValueError("can be either 'drop' or 'passthrough'")
        return data


class DatetimeSchema(BaseModel):
    type: str
    ds: Optional[str] = None
    remainder: str = "passthrough"
    replace: bool = False
    subset: Optional[Union[str, List[str]]] = None

    @validator("remainder")
    def check_remainder(cls, data):
        if data not in ("drop", "passthrough"):
            raise ValueError("can be either 'drop' or 'passthrough'")
        return data

    @validator("subset")
    def check_subset(cls, data):
        if isinstance(data, str):
            data = [x.strip() for x in data.split(",")]
        if (not isinstance(data, list)) or (not all(isinstance(x, str) for x in data)):
            raise ValueError(
                "must be a list of strings or a string of comma-separated values."
            )
        return data


class CyclicalSchema(BaseModel):
    type: str
    seasonality: str
    ds: Optional[str] = None
    period: Optional[float] = None
    fourier_order: Optional[int] = None
    remainder: str = "passthrough"
    replace: bool = False

    @validator("remainder")
    def check_remainder(cls, data):
        if data not in ("drop", "passthrough"):
            raise ValueError("can be either 'drop' or 'passthrough'")
        return data


class LinearSchema(BaseModel):
    type: str
    feature: str
    as_filter: bool = False
    include_bias: bool = False


class SplineSchema(BaseModel):
    type: str
    feature: str
    n_knots: Optional[int] = 5
    degree: Optional[int] = 3
    strategy: Optional[Union[str, List]] = "uniform"
    extrapolation: Optional[str] = "constant"
    include_bias: bool = False

    @validator("strategy")
    def check_strategy(cls, data):
        if (data is None) or (data in ("uniform", "quantile")):
            return data
        if isinstance(data, list) and all(isinstance(x, (int, float)) for x in data):
            return data
        else:
            raise ValueError(
                "can be one of 'uniform' or 'quantile', or array-like of numbers"
            )

    @validator("extrapolation")
    def check_extrapolation(cls, data):
        if (data is not None) and (
            data not in ("error", "constant", "linear", "continue")
        ):
            raise ValueError("can be one of 'error', 'constant', 'linear', 'continue'")
        return data


class CategoricalSchema(BaseModel):
    type: str
    feature: str
    max_n_categories: Optional[int] = None
    stratify_by: Optional[Union[str, List[str]]] = None
    excluded_categories: Optional[Union[str, List[str]]] = None
    unknown_value: Optional[int] = None
    min_samples_leaf: int = 1
    max_features: Union[str, int, float] = "auto"
    random_state: Optional[int] = None
    encode_as: str = "onehot"

    @validator("stratify_by", "excluded_categories")
    def check_lists(cls, data):
        if isinstance(data, str):
            data = [x.strip() for x in data.split(",")]
        if (not isinstance(data, list)) or (not all(isinstance(x, str) for x in data)):
            raise ValueError(
                "must be a list of strings or a string of comma-separated values."
            )
        return data

    @validator("max_features")
    def check_max_features(cls, data):
        if data in ("auto", "sqrt", "log2"):
            return data
        if data.replace(".", "", 1).isdigit():
            if isinstance(data, numbers.Integral):
                return data
            elif (data > 0) and (data <= 1):
                return data
        raise ValueError(
            "can be int, float between 0 and 1, or one of 'auto', 'sqrt', 'log2'"
        )

    @validator("encode_as")
    def check_encode_as(cls, data):
        if data not in ("onehot", "ordinal"):
            raise ValueError("can be either 'onehot' or 'ordinal'")
        return data



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from collections import OrderedDict

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin, clone
from sklearn.exceptions import NotFittedError
from sklearn.utils.validation import check_is_fitted

from ..utils import tensor_product
from ._encoders import CategoricalEncoder, IdentityEncoder, SplineEncoder

# ------------------------------------------------------------------------------------
# Encode pairwise categorical data interactions
#
# All encoders generate numpy arrays
# ------------------------------------------------------------------------------------


class ICatEncoder(TransformerMixin, BaseEstimator):
    """Encode the interaction between two categorical features.

     Interactions are always pairwise and always between encoders (and not features).

    Args:
        encoder_left (CategoricalEncoder): The encoder for the first of the two features.
        encoder_right (CategoricalEncoder): The encoder for the second of the two features.

    Raises:
        ValueError: If any of the two encoders is not a `CategoricalEncoder`.
        ValueError: If the two encoders do not have the same `encode_as` parameter.

    Note:
        Both encoders should have the same `encode_as` parameter.
        If one or both of the encoders is already fitted, it will not be re-fitted during
        `fit` or `fit_transform`.
    """

    def __init__(
        self, encoder_left: CategoricalEncoder, encoder_right: CategoricalEncoder
    ):
        if (not isinstance(encoder_left, CategoricalEncoder)) or (
            not isinstance(encoder_right, CategoricalEncoder)
        ):
            raise ValueError(
                "This pairwise interaction encoder expects `CategoricalEncoder` encoders"
            )
        if encoder_left.encode_as != encoder_right.encode_as:
            raise ValueError(
                "Both encoders should have the same `encode_as` parameter."
            )

        self.encoder_left = encoder_left
        self.encoder_right = encoder_right
        self.encode_as_ = encoder_left.encode_as

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            ICatEncoder: Fitted encoder.
        """
        for encoder in (self.encoder_left, self.encoder_right):
            try:
                check_is_fitted(encoder, "fitted_")
            except NotFittedError:
                encoder.fit(X, y)

        self.n_features_out_ = (
            self.encoder_left.n_features_out_ * self.encoder_right.n_features_out_
        )
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Returns:
            numpy array: The matrix of interaction features as a numpy array.
        """
        check_is_fitted(self, "fitted_")

        X_left = self.encoder_left.transform(X)
        X_right = self.encoder_right.transform(X)

        if self.encode_as_ == "onehot":
            return tensor_product(X_left, X_right)
        else:
            X_left = X_left.astype(str, copy=False)
            X_right = X_right.astype(str, copy=False)
            X_left = np.core.defchararray.add(X_left, np.array([":"]))
            return np.core.defchararray.add(X_left, X_right)


# ------------------------------------------------------------------------------------
# Encode pairwise interactions between numerical features
# ------------------------------------------------------------------------------------


class ISplineEncoder(TransformerMixin, BaseEstimator):
    """Encode the interaction between two spline-encoded numerical features.

    Args:
        encoder_left (SplineEncoder): The encoder for the first of the two features.
        encoder_right (SplineEncoder): The encoder for the second of the two features.

    Raises:
        ValueError:If any of the two encoders is not a `SplineEncoder`.

    Note:
        If one or both of the encoders is already fitted, it will not be re-fitted
        during `fit` or `fit_transform`.
    """

    def __init__(self, encoder_left: SplineEncoder, encoder_right: SplineEncoder):
        if (not isinstance(encoder_left, SplineEncoder)) or (
            not isinstance(encoder_right, SplineEncoder)
        ):
            raise ValueError(
                "This pairwise interaction encoder expects `SplineEncoder` encoders"
            )

        self.encoder_left = encoder_left
        self.encoder_right = encoder_right

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            ISplineEncoder: Fitted encoder.
        """
        for encoder in (self.encoder_left, self.encoder_right):
            try:
                check_is_fitted(encoder, "fitted_")
            except NotFittedError:
                encoder.fit(X)

        self.n_features_out_ = (
            self.encoder_left.n_features_out_ * self.encoder_right.n_features_out_
        )
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Returns:
            numpy array: The matrix of interaction features as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X_left = self.encoder_left.transform(X)
        X_right = self.encoder_right.transform(X)
        return tensor_product(X_left, X_right)


class ProductEncoder(TransformerMixin, BaseEstimator):
    """Encode the interaction between two linear numerical features.

    Args:
        encoder_left (IdentityEncoder): The encoder for the first of the two features.
        encoder_right (IdentityEncoder): The encoder for the second of the two features.

    Raises:
        ValueError: If any of the two encoders is not an `IdentityEncoder`.

    Note:
        If one or both of the encoders is already fitted, it will not be re-fitted during
        `fit` or `fit_transform`.
    """

    def __init__(self, encoder_left: IdentityEncoder, encoder_right: IdentityEncoder):
        if (not isinstance(encoder_left, IdentityEncoder)) or (
            not isinstance(encoder_right, IdentityEncoder)
        ):
            raise ValueError(
                "This pairwise interaction encoder expects `IdentityEncoder` encoders"
            )

        self.encoder_left = encoder_left
        self.encoder_right = encoder_right

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Raises:
            ValueError: If any of the two encoders is not a single-feature encoder.

        Returns:
            ProductEncoder: Fitted encoder.
        """
        for encoder in (self.encoder_left, self.encoder_right):
            try:
                check_is_fitted(encoder, "fitted_")
            except NotFittedError:
                encoder.fit(X)
            finally:
                if encoder.n_features_out_ > 1:
                    raise ValueError(
                        "This pairwise interaction encoder supports only single-feature encoders"
                    )

        self.n_features_out_ = 1
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Returns:
            numpy array: The matrix of interaction features as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X_left = self.encoder_left.transform(X)
        X_right = self.encoder_right.transform(X)
        return np.multiply(X_left, X_right)


# ------------------------------------------------------------------------------------
# Encode pairwise interactions of one numerical and one categorical feature
# ------------------------------------------------------------------------------------


class ICatLinearEncoder(TransformerMixin, BaseEstimator):
    """Encode the interaction between one categorical and one linear numerical feature.

    Args:
        encoder_cat (CategoricalEncoder): The encoder for the categorical feature.
            It must encode features in an one-hot form.
        encoder_num (IdentityEncoder): The encoder for the numerical feature.

    Raises:
        ValueError: If `encoder_cat` is not  a `CategoricalEncoder`.
        ValueError: If `encoder_num` is not  an `IdentityEncoder`.
        ValueError: If `encoder_cat` is not encoded as one-hot.

    Note:
        If one or both of the encoders is already fitted, it will not be re-fitted
        during `fit` or `fit_transform`.
    """

    def __init__(
        self, *, encoder_cat: CategoricalEncoder, encoder_num: IdentityEncoder
    ):
        if not isinstance(encoder_cat, CategoricalEncoder):
            raise ValueError("`encoder_cat` must be a CategoricalEncoder")

        if encoder_cat.encode_as != "onehot":
            raise ValueError(
                "This encoder supports only one-hot encoding of the "
                "categorical feature"
            )

        if not isinstance(encoder_num, IdentityEncoder):
            raise ValueError("`encoder_num` must be an IdentityEncoder")

        self.encoder_cat = encoder_cat
        self.encoder_num = encoder_num

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            ICatLinearEncoder: Fitted encoder.
        """
        try:
            check_is_fitted(self.encoder_cat, "fitted_")
        except NotFittedError:
            self.encoder_cat.fit(X, y)

        try:
            check_is_fitted(self.encoder_num, "fitted_")
        except NotFittedError:
            self.encoder_num.fit(X)

        self.n_features_out_ = (
            self.encoder_cat.n_features_out_ * self.encoder_num.n_features_out_
        )
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Returns:
            numpy array: The matrix of interaction features as a numpy array.
        """
        check_is_fitted(self, "fitted_")
        X_cat = self.encoder_cat.transform(X)
        X_num = self.encoder_num.transform(X)
        # guard for single category
        if X_cat.shape[1] == 1:
            return X_num
        else:
            return tensor_product(X_cat, X_num)


class ICatSplineEncoder(TransformerMixin, BaseEstimator):
    """Encode the interaction between one categorical and one spline-encoded
    numerical feature.

    Args:
        encoder_cat (CategoricalEncoder): The encoder for the categorical
            feature. It must encode features in an one-hot form.
        encoder_num (SplineEncoder): The encoder for the numerical feature.

    Raises:
        ValueError: If `encoder_cat` is not  a `CategoricalEncoder`.
        ValueError: If `encoder_num` is not  a `SplineEncoder`.
        ValueError: If `encoder_cat` is not encoded as one-hot.

    Note:
        If the categorical encoder is already fitted, it will not be re-fitted
        during `fit` or `fit_transform`.
        The numerical encoder will always be (re)fitted (one encoder per level
        of categorical feature.
    """

    def __init__(
        self,
        *,
        encoder_cat: CategoricalEncoder,
        encoder_num: SplineEncoder,
    ):
        if not isinstance(encoder_cat, CategoricalEncoder):
            raise ValueError("`encoder_cat` must be a CategoricalEncoder")

        if encoder_cat.encode_as != "onehot":
            raise ValueError(
                "This encoder supports only one-hot encoding of the "
                "categorical feature"
            )

        if not isinstance(encoder_num, SplineEncoder):
            raise ValueError("`encoder_num` must be a SplineEncoder")

        self.encoder_cat = encoder_cat
        self.encoder_num = encoder_num

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the encoder on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            ICatSplineEncoder: Fitted encoder.
        """
        try:
            check_is_fitted(self.encoder_cat, "fitted_")
        except NotFittedError:
            self.encoder_cat.fit(X, y)

        encoders = OrderedDict({})
        cat_features = pd.DataFrame(data=self.encoder_cat.transform(X), index=X.index)

        for i, col in enumerate(cat_features.columns):
            mask = cat_features[col] == 1
            enc = clone(self.encoder_num)
            try:
                encoders[i] = enc.fit(X.loc[mask])
            except ValueError:
                encoders[i] = enc

        self.num_encoders_ = encoders
        self.n_features_out_ = (
            len(encoders) * next(iter(encoders.values())).n_features_out_
        )
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the encoder.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input
                dataframe.

        Returns:
            numpy array : The matrix of interaction features as a numpy array.
        """
        check_is_fitted(self, "fitted_")

        out = None
        cat_features = pd.DataFrame(data=self.encoder_cat.transform(X), index=X.index)

        for i, encoder in self.num_encoders_.items():
            mask = cat_features.loc[:, i] == 1
            subset = X.loc[mask]
            if subset.empty or (not encoder.fitted_):
                trf = pd.DataFrame(
                    data=np.zeros((X.shape[0], encoder.n_features_out_)), index=X.index
                )
            else:
                trf = pd.DataFrame(
                    data=encoder.transform(subset), index=subset.index
                ).reindex(X.index, fill_value=0)
            out = pd.concat((out, trf), axis=1)

        out = np.array(out)
        return out



import os
from setuptools import find_packages, setup

from feature_encoders import __version__


here = os.path.abspath(os.path.dirname(__file__))


def dev_extras_require():
    extras = [
        "flake8 >= 3.9.2",
        "flake8-docstrings >= 1.6.0",
        "black >= 21.7b0",
        "pytest >= 6.2.5",
    ]
    return extras


def docs_extras_require():
    extras = [
        'Sphinx >= 3.0.0',  # Force RTD to use >= 3.0.0
        'docutils < 0.18',
        "nbsphinx >= 0.8.7",
        'pylons-sphinx-themes >= 1.0.8',  # Ethical Ads
        "myst-parse >= 0.15.2"
    ]
    return extras


# get the dependencies and installs
with open("requirements.txt", "r", encoding="utf-8") as f:
    requires = []
    for line in f:
        req = line.split("#", 1)[0].strip()
        if req and not req.startswith("--"):
            requires.append(req)


# Get the long description from the README file
with open(os.path.join(here, "README.md"), encoding="utf-8") as f:
    readme = f.read()



configuration = {
    "name": "feature-encoders",
    "version": __version__,
    "python_requires=": ">=3.7",
    "description": (
        "A library for encoding features and their pairwise interactions."
    ),
    "long_description": readme,
    "long_description_content_type": "text/markdown",
    "classifiers": [
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Operating System :: MacOS",
    ],
    "keywords": "feature encoding interactions",
    "url": "https://github.com/hebes-io/feature-encoders",
    "maintainer": "Sotiris Papadelis",
    "maintainer_email": "spapadel@gmail.com",
    "license": "Apache License, Version 2.0",
    "packages": find_packages(),
    "install_requires": requires,
    "ext_modules": [],
    "cmdclass": {},
    "include_package_data": True,
    "package_data": {'feature_encoders': ['config/features/*.yaml', 'config/models/*.yaml']},
    "extras_require": {"dev": dev_extras_require(), "docs": docs_extras_require()},
}

setup(**configuration)



# -*- coding: utf-8 -*-
# Copyright (c) Hebes Intelligence Private Company

# This source code is licensed under the Apache License, Version 2.0 found in the
# LICENSE file in the root directory of this source tree.

from datetime import datetime

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.preprocessing import MinMaxScaler
from sklearn.utils.validation import check_is_fitted

from ..utils import as_list, check_X, get_datetime_data

#####################################################################################
# Add new features
#
# All feature generators generate pandas DataFrames
#####################################################################################


class TrendFeatures(TransformerMixin, BaseEstimator):
    """Generate linear time trend features.

    Args:
        ds (str, optional): The name of the input dataframe's column that contains
            datetime information. If None, it is assumed that the datetime information
            is provided by the input dataframe's index. Defaults to None.
        name (str, optional): The name of the generated dataframe's column. Defaults to 'growth'.
        remainder ({'drop', 'passthrough'}, optional): By specifying ``remainder='passthrough'``,
            all the remaining columns of the input dataset will be automatically passed through
            (concatenated with the output of the transformer), otherwise, they will be dropped.
            Defaults to "passthrough".
        replace (bool, optional): Specifies whether replacing an existing column with the same
            name is allowed (applicable when `remainder=passthrough`). Defaults to False.

    Raises:
        ValueError: If ``remainder`` is neither 'drop' nor 'passthrough'.
    """

    def __init__(self, ds=None, name="growth", remainder="passthrough", replace=False):
        if remainder not in ("passthrough", "drop"):
            raise ValueError("Parameter `remainder` should be 'passthrough' or 'drop'")

        self.ds = ds
        self.name = name
        self.remainder = remainder
        self.replace = replace

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the feature generator on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            TrendFeatures: Fitted encoder.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
        """
        X = check_X(X)
        dates = X.index.to_frame() if self.ds is None else X[[self.ds]]
        self.t_scaler_ = MinMaxScaler().fit(dates)
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the feature generator.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If common columns are found and ``replace=False``.

        Returns:
            pandas.DataFrame: The transformed dataframe.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X)
        dates = X.index.to_frame() if self.ds is None else X[[self.ds]]

        out = pd.DataFrame(
            data=self.t_scaler_.transform(dates),
            columns=[self.name],
            index=X.index,
        )

        if self.remainder == "passthrough":
            if (self.name in X.columns) and not self.replace:
                raise ValueError(f"Found common column name: {self.name}")
            elif self.name in X.columns:
                X = X.drop(self.name, axis=1)
            out = pd.concat((X, out), axis=1)

        return out


class DatetimeFeatures(TransformerMixin, BaseEstimator):
    """Generate date and time features.

    Args:
        ds (str, optional): The name of the input dataframe's column that contains
            datetime information. If None, it is assumed that the datetime information
            is provided by the input dataframe's index. Defaults to None.
        remainder ({'drop', 'passthrough'}, optional): By specifying ``remainder='passthrough'``,
            all the remaining columns of the input dataset will be automatically passed through
            (concatenated with the output of the transformer), otherwise, they will be dropped.
            Defaults to "passthrough".
        replace (bool, optional): Specifies whether replacing an existing column with the same
            name is allowed (applicable when `remainder=passthrough`). Defaults to False.
        subset (str or list of str, optional): The names of the features to generate. If
            None, all features will be produced: 'month', 'week', 'dayofyear', 'dayofweek',
            'hour', 'hourofweek'. The last 2 features are generated only if the timestep of
            the input's `ds` (or index if `ds` is None) is smaller than `pandas.Timedelta(days=1)`.
            Defaults to None.

    Raises:
        ValueError: If ``remainder`` is neither 'drop' nor 'passthrough'.
    """

    def __init__(self, ds=None, remainder="passthrough", replace=False, subset=None):
        if remainder not in ("passthrough", "drop"):
            raise ValueError('Parameter "remainder" should be "passthrough" or "drop"')

        self.ds = ds
        self.remainder = remainder
        self.replace = replace
        self.subset = subset

    def _get_all_attributes(self, dt_column):
        attr = ["month", "week", "dayofyear", "dayofweek"]

        dt = dt_column.diff()
        time_step = dt.iloc[dt.values.nonzero()[0]].min()
        if time_step < pd.Timedelta(days=1):
            attr = attr + ["hour", "hourofweek"]

        if self.subset is not None:
            attr = [i for i in attr if i in as_list(self.subset)]
        return attr

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the feature generator on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            DatetimeFeatures: Fitted encoder.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
        """

        X = check_X(X)
        dt_column = get_datetime_data(X, col_name=self.ds)
        self.attr_ = self._get_all_attributes(dt_column)
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the feature generator.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If common columns are found and ``replace=False``.

        Returns:
            pandas.DataFrame: The transformed dataframe.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X)
        dt_column = get_datetime_data(X, col_name=self.ds)

        out = {}
        for n in self.attr_:
            if n == "week":
                out[n] = (
                    dt_column.dt.isocalendar().week.astype(dt_column.dt.day.dtype)
                    if hasattr(dt_column.dt, "isocalendar")
                    else dt_column.dt.week
                )
            elif n == "hourofweek":
                out[n] = None
            else:
                out[n] = getattr(dt_column.dt, n)

        if "hourofweek" in out:
            out["hourofweek"] = 24 * out.get(
                "dayofweek", dt_column.dt.dayofweek
            ) + out.get("hour", dt_column.dt.hour)

        out = pd.DataFrame.from_dict(out)

        if self.remainder == "passthrough":
            common = list(set(X.columns) & set(out.columns))
            if common and not self.replace:
                raise ValueError(f"Found common column names {common}")
            elif common:
                X = X.drop(common, axis=1)
            out = pd.concat((X, out), axis=1)

        return out


class CyclicalFeatures(TransformerMixin, BaseEstimator):
    """Create cyclical (seasonal) features as fourier terms.

    Args:
        seasonality (str): The name of the seasonality. The feature generator can provide
            default values for ``period`` and ``fourier_order`` if ``seasonality`` is one
            of 'daily', 'weekly' or 'yearly'.
        ds (str, optional): The name of the input dataframe's column that contains
            datetime information. If None, it is assumed that the datetime information
            is provided by the input dataframe's index. Defaults to None.
        period (float, optional): Number of days in one period. Defaults to None.
        fourier_order (int, optional): Number of Fourier components to use. Defaults to None.
        remainder ({'drop', 'passthrough'}, optional): By specifying ``remainder='passthrough'``,
            all the remaining columns of the input dataset will be automatically passed through
            (concatenated with the output of the transformer), otherwise, they will be dropped.
            Defaults to "passthrough".
        replace (bool, optional): Specifies whether replacing an existing column with the same
            name is allowed (applicable when `remainder=passthrough`). Defaults to False.

    Raises:
        ValueError: If ``remainder`` is neither 'drop' nor 'passthrough'.
    """

    def __init__(
        self,
        *,
        seasonality,
        ds=None,
        period=None,
        fourier_order=None,
        remainder="passthrough",
        replace=False,
    ):
        if remainder not in ("passthrough", "drop"):
            raise ValueError('Parameter "remainder" should be "passthrough" or "drop"')

        self.seasonality = seasonality
        self.ds = ds
        self.period = period
        self.fourier_order = fourier_order
        self.remainder = remainder
        self.replace = replace

    @staticmethod
    def _fourier_series(dates, period, order):
        # convert to days since epoch
        t = np.array(
            (dates - datetime(2000, 1, 1)).dt.total_seconds().astype(np.float64)
        ) / (3600 * 24.0)

        return np.column_stack(
            [
                fun((2.0 * (i + 1) * np.pi * t / period))
                for i in range(order)
                for fun in (np.sin, np.cos)
            ]
        )

    def fit(self, X: pd.DataFrame, y=None):
        """Fit the feature generator on the available data.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.
            y (None, optional): Ignored. Defaults to None.

        Returns:
            CyclicalFeatures: Fitted encoder.

        Raises:
            ValueError: If either ``period`` or ``fourier_order`` is not provided, but
                ``seasonality`` is not one of 'daily', 'weekly' or 'yearly'.
        """
        if self.seasonality not in ["daily", "weekly", "yearly"]:
            if (self.period is None) or (self.fourier_order is None):
                raise ValueError(
                    "When adding custom seasonalities, values for "
                    "`period` and `fourier_order` must be specified."
                )
        if self.seasonality in ["daily", "weekly", "yearly"]:
            if self.period is None:
                self.period = (
                    1
                    if self.seasonality == "daily"
                    else 7
                    if self.seasonality == "weekly"
                    else 365.25
                )
            if self.fourier_order is None:
                self.fourier_order = (
                    4
                    if self.seasonality == "daily"
                    else 3
                    if self.seasonality == "weekly"
                    else 6
                )
        self.fitted_ = True
        return self

    def transform(self, X: pd.DataFrame):
        """Apply the feature generator.

        Args:
            X (pandas.DataFrame of shape (n_samples, n_features)): The input dataframe.

        Raises:
            ValueError: If the input data does not pass the checks of `utils.check_X`.
            ValueError: If common columns are found and ``replace=False``.

        Returns:
            pandas.DataFrame: The transformed dataframe.
        """
        check_is_fitted(self, "fitted_")
        X = check_X(X)
        dt_column = get_datetime_data(X, col_name=self.ds)
        out = self._fourier_series(dt_column, self.period, self.fourier_order)
        out = pd.DataFrame(
            data=out,
            index=X.index,
            columns=[
                f"{self.seasonality}_delim_{i}" for i in range(2 * self.fourier_order)
            ],
        )

        if self.remainder == "passthrough":
            common = list(set(X.columns) & set(out.columns))
            if common and not self.replace:
                raise ValueError(f"Found common column names {common}")
            elif common:
                X = X.drop(common, axis=1)
            out = pd.concat((X, out), axis=1)

        return out
